/* 
   @author Korniloff E.
     2007
*/


#include "chess.h"
#include "bitmask.h"


score_node score_tree[MAX_PLY];


#ifdef DEBUG
  int __error;
  int tempBuf[8][8];
#endif

const int PIN  = 6;
const int XRAY = 4;
//const int HUNG = 6;      //[] = {0,10,3,3,5,10,20};
const int HUNG[] = {0,6,3,3,5,10,20};


const int PBLOCK         = 1; //   
const int PBLOCK_OP[]    = {1,1,2,3,3,2,1,1}; //
const int PBLOCK_PASSED  = 4; //

const int PDOUBLE = 6;  //
const int PISOL   = 8;  //
const int PBACK   = 1;  //

const int RAPP  = 6; //  
const int RHOF  = 4; //rook half open file
const int ROF   = 6; //rook open file
const int R7R   = 8; // rook 7 rank
const int R7R_ZOOM = 12; // ---   -    

const int NNK = 4;  //  knight near by king

int value[7] = {0,VALUE_P,VALUE_N,VALUE_B,VALUE_R,VALUE_Q,VALUE_K};
int score_table[2][8][64];
extern int history[2][8][64];
int FullPawnAtk = 0;

 
const int PASSED_PAWN_ROW[2][8] =
{
  {0,76,30,20,15,10,5,0},
  {0,5,10,15,20,30,76,0}
 
};
 

const int HALF_PASSED_PAWN_ROW[2][8] =
{
  {0,20,16,12,8,4,0,0},
  {0,0,4,8,12,16,20,0}
};
const int rank[2][8] =
{
  {7,6,5,4,3,2,1,0},
  {0,1,2,3,4,5,6,7}
};
const int main_line[8] =  {1,1,2,3,3,2,1,1};
const int knight_forpost_column[8] = {0,1,2,4,4,2,1,0};

const int comp_value[] = {0,1,3,3,5,10,20};
#define CNTRL_PAWN   (1<<14)
#define CNTRL_KNIGHT (1<<13)
#define CNTRL_BISHOP (1<<12)
#define CNTRL_ROOK   (1<<11)
#define CNTRL_QUEEN  (1<<10)
#define CNTRL_KING   (1<<9)
const int piece_cntrl[] = {0,CNTRL_PAWN,CNTRL_KNIGHT,CNTRL_BISHOP,
                           CNTRL_ROOK,CNTRL_QUEEN,CNTRL_KING};

const int less_piece_cntrl[] = 
{
  0, // 
  0, // for pawn
  CNTRL_PAWN, //KNIGHT
  CNTRL_PAWN, // BISHOP
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT, //ROOK
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT|CNTRL_ROOK, //QUEEN
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT|CNTRL_ROOK|CNTRL_QUEEN //KING
};
const int less_or_like_piece_cntrl[] = 
{
  0, // 
  CNTRL_PAWN, // for pawn
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT, //KNIGHT
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT, // BISHOP
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT|CNTRL_ROOK, //ROOK
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT|CNTRL_ROOK|CNTRL_QUEEN, //QUEEN
  CNTRL_PAWN|CNTRL_BISHOP|CNTRL_KNIGHT|CNTRL_ROOK|CNTRL_QUEEN //KING
};
const int pawn_start_row[8] = {NEUTRAL, BLACK, NEUTRAL, NEUTRAL,
                               NEUTRAL, NEUTRAL,WHITE,NEUTRAL};
#define PAWN_START_POS(sq,c)  ( pawn_start_row[ROW(sq)]==c )
const int opponent_board[] = {WHITE,WHITE,WHITE,WHITE,
                              BLACK,BLACK,BLACK,BLACK};
const int main_diag[64] = {
  2,1,0,0,0,0,1,2,
  1,2,1,0,0,1,2,1,
  0,1,2,1,1,2,1,0,
  0,0,1,2,2,1,0,0,
  0,0,1,2,2,1,0,0,
  0,1,2,1,1,2,1,0,
  1,2,1,0,0,1,2,1,
  2,1,0,0,0,0,1,2

};



struct Atk_Type{
  int bit_atk[64];


} p_attack[2];



int max(int v1, int v2) {  return v1 > v2 ? v1 : v2; }
int min(int v1, int v2) {  return v1 < v2 ? v1 : v2; }

//forward
double StageKnightBishop(int c);
double Stage(int c);
//#undef abs
/*
int abs(int v){
  return v >= 0 ? v : (-v);
}
*/
int Distance(int v1, int v2){

  return max(abs(ROW(v1)-ROW(v2)),
             abs(COLUMN(v1)-COLUMN(v2)));
}

int Taxi(int sq0, int sq1)
{
   return abs(COLUMN(sq0)-COLUMN(sq1)) + abs(ROW(sq0)-ROW(sq1));
}




/*
      (0..1)
*/
double Stage(int c){
   const int EMTL = 14 + 2*6 + 2*4 + 2*3;   // 42
    int c1 = c^1;
    int *v = g.cnt[c1],
        emtl = v[QUEEN]*14 +
               v[ROOK]*6 +
               v[BISHOP]*4 +
               v[KNIGHT]*3;

    if(v[QUEEN]){
      int cnt = v[ROOK] + v[BISHOP] + v[KNIGHT];
      if(cnt){
         emtl += 1;
         if( cnt > 1 ) emtl += 1;
      }
    }

    if(emtl > EMTL) emtl = EMTL;
    return (double)emtl / EMTL;
}

/*
     
  4 - 
  3 - 
  2 -   
  < 2  -  0 
*/
double StageKnightBishop(int c){
   const int EMTL =  2*4 + 2*3;   // 14
   int *v = g.cnt[c ^ 1],
        cnt  =  v[BISHOP] + v[KNIGHT],
        emtl = v[BISHOP]*4 +
               v[KNIGHT]*3;
    if( cnt <= 1 ) return 0;
    if( cnt == 2 ) emtl -= 2;

    if(emtl > EMTL) emtl = EMTL;
    return (double)emtl / EMTL;

}

void InitEvaluate(void){
 void  GenScoreTable(int p, int tbl[64], int c, int max, int op_king_sq);
 int p,c,j;
 const int blackPawnScore[64] =
   { 0, 0, 0, 0, 0, 0, 0, 0,
     4, 4, 4, 0, 0, 4, 4, 4,
     6, 8, 2,10,10, 2, 8, 6,
     6, 8,12,16,16,12, 8, 6,
     8,12,16,24,24,16,12, 8,
    12,16,24,32,32,24,16,12,
    12,16,24,32,32,24,16,12,
     0, 0, 0, 0, 0, 0, 0, 0};



 for(c = WHITE; c <= BLACK; c++){
   for(p = PAWN; p <= KING; p++)
   {
     const int kf[] = {0,32,  // pawn
                     32,//36,  //knight
                     32,//30,  //bishop
                     32,  //rook
                     32,//30  //queen
                     36   //king
                     };


     GenScoreTable(p,score_table[c][p],c, kf[p], g.kingSq[c^1]);

   }
 }
 score_table[BLACK][BISHOP][B7] += 2;
 score_table[BLACK][BISHOP][G7] += 2;
 score_table[WHITE][BISHOP][B2] += 2;
 score_table[WHITE][BISHOP][G2] += 2;

 for(j = 0; j < 64; j++){
    score_table[BLACK][PAWN][j] =  blackPawnScore[j];
    score_table[WHITE][PAWN][j] =  blackPawnScore[63-j];
 }
/*
 for(c = WHITE; c <= BLACK; c++){
    const king_file[] = {0,0,3,4,4,3,0,0};
    double stage = Stage(c);
    int j;


    if(g.cnt[c][PAWN]==0) stage = 0; //    

    for(j = 0; j < 64; j++){
      double tmp = score_table[c][KING][j];
      if(stage > 0.5) tmp += king_file[COLUMN(j)];

      score_table[c][KING][j] =  -tmp * stage  +  tmp * (1.0 - stage);
    }
 }
*/

 FullPawnAtk = !((g.cnt[WHITE][QUEEN]>0)  &&  (g.cnt[BLACK][QUEEN]>0));



 // recalc all step-score
 do{
   int c,i,sq,p;
    for(c = WHITE; c <= BLACK; c++)
    {
      g.st_score[c] = 0;
      for(i = g.start[c]; i <= g.stop[c]; i++)
       if( (sq = g.list[i]) != -1 )
       {
         p = g.pos[sq]; 
         g.st_score[c] +=  score_table[c][p][sq];
       }  
    }   
 }while(0);

  /*
 do{
    void PrintArray(FILE*f,int tbl[64], int p, int c);
    FILE *f = fopen("sctable.txt","w");
    int p,c;

    for(c = WHITE; c <= BLACK; c++)
     for(p = PAWN; p <= KING; p++)
       PrintArray(f, score_table[c][p], p, c);

    fclose(f);

 }while(0);
    */
}


int CanUseNullMove(void){

   
  void GenerateAttackes(void);
  int c,i,sq,p,dP,  *b_atk1,
      cntSweep[2]={0,0},cntNotBlockedPawn = 0;

    GenerateAttackes();
    for(c = WHITE; c <= BLACK; c++){
      dP = c==WHITE ? -8:8;
     // b_atk0 = p_attack[c].bit_atk;
      b_atk1 = p_attack[c^1].bit_atk;
      for(i = g.start[c]; i <= g.stop[c]; i++)
       if( (sq = g.list[i]) != -1 ){


         p = g.pos[sq];
         if(SWEEP(p))  cntSweep[c]++;
         if(p == PAWN){
            if( g.color[sq+dP]==(c^1) ||  (b_atk1[sq+dP] & CNTRL_PAWN) ||
                g.pos[sq+dP] == PAWN
            )
            {

            }else cntNotBlockedPawn++;
         }
       }
    }

   if( cntSweep[WHITE] >= 1  && cntSweep[BLACK] >= 1 &&
       cntNotBlockedPawn >=  3  &&  Stage(WHITE) > 0.3 && Stage(BLACK) > 0.3)
       return 1;
   else return 0;
}


const int a1h8[64] =
{
  1,2,3,4,5,6,7,8,
  2,3,4,5,6,7,8,9,
  3,4,5,6,7,8,9,10,
  4,5,6,7,8,9,10,11,
  5,6,7,8,9,10,11,12,
  6,7,8,9,10,11,12,13,
  7,8,9,10,11,12,13,14,
  8,9,10,11,12,13,14,15
};

const int a8h1[64] =
{
   8,9,10,11,12,13,14,
   7,8,9,10,11,12,13,14,
   6,7,8,9,10,11,12,13,
   5,6,7,8,9,10,11,12,
   4,5,6,7,8,9,10,11,
   3,4,5,6,7,8,9,10,
   2,3,4,5,6,7,8,9,
   1,2,3,4,5,6,7,8
};



int Evaluate(int alpha, int beta){
                 // max(350 - ply, 320);
  const int MARGIN =  //max(300 - ply, 250);
                  //max(250 - ply, 150);
      ((g.mtl[WHITE]>VALUE_K) & (g.mtl[WHITE]>VALUE_K)) ?
         350 - ply: 100000;

  int s[2];
  static HashKey save_key;
  static int save_side,save_alpha,save_beta,save_score;





  if(g.key == save_key && save_side == g.side &&
     save_alpha <= alpha && save_beta >= beta)
      return save_score;



  s[WHITE] = g.mtl[WHITE] + g.st_score[WHITE];
  s[BLACK] = g.mtl[BLACK] + g.st_score[BLACK];

  if( (g.cnt[WHITE][PAWN]  | g.cnt[BLACK][PAWN])==0  &&
      (g.cnt[WHITE][QUEEN] | g.cnt[BLACK][QUEEN])==0 &&
      (g.cnt[WHITE][ROOK ] | g.cnt[BLACK][ROOK])==0  &&
       g.cnt[WHITE][BISHOP] * 4   +   g.cnt[WHITE][KNIGHT] * 3   <  4+3 &&
       g.cnt[BLACK][BISHOP] * 4   +   g.cnt[BLACK][KNIGHT] * 3   <  4+3 )
       return 0;

  if(s[g.side] - s[g.xside] + MARGIN <= alpha  ||
     s[g.side] - s[g.xside] - MARGIN >= beta)
     return s[g.side] - s[g.xside];


  do{
    void GenerateAttackes(void);
    int Trapped(int sq, int p, int c, int* a0, int* a1);
    int  Mobile_Pin_XRay(int sq, int p, int c, int* a0, int* a1);
    int KingSFTY(int c);
    int c,i,sq,p,dP,*b_atk0, *b_atk1,c1,hungCnt,isolCnt,score;//,st_score;
    double stage,__stage;

    GenerateAttackes();
    for(c = WHITE; c <= BLACK; c++){
      double stageNB = StageKnightBishop(c);
      double __stageNB = 1.0 - stageNB;
      int cntDoublePawn = 0;
      int cntPassed = 0;
      int rook7RankCnt = 0;
      int passedPawnColMask = 0;
      int hasQueen = g.cnt[c][QUEEN];
      int *pawn_column_cnt = g.pawn_column_cnt[c];
      int opKingSq = g.kingSq[c^1];

      dP = c==WHITE ? -8:8;
      c1 = c^1;
      b_atk0 = p_attack[c].bit_atk;
      b_atk1 = p_attack[c1].bit_atk;
      hungCnt = score = isolCnt = 0;
      //st_score = 0;
      stage = Stage(c);
      __stage = 1.0 - stage;
      
      
      
      for(i = g.start[c]; i <= g.stop[c]; i++)
       if( (sq = g.list[i]) != -1 ){
         int b0,b1;
         int x = COLUMN(sq), y = ROW(sq);


         p = g.pos[sq];
         //score += score_table[c][p][sq];
         b0 = b_atk0[sq]; // attacks on square
         b1 = b_atk1[sq]; // opponent attacks 


         // 
         if(b0==0) score -= 1; //  
         else if(less_or_like_piece_cntrl[p] & b0)
           score += 1; //    
          
         if(b1) //
         {
           score -= b1 & 15; // -          
           if(b0==0  ||  (less_piece_cntrl[p] & b1) ){
              //
              score -= HUNG[p];
              ++hungCnt;
              if(hungCnt==2) score -= 4;
           }else if(p==PAWN && (b0 & CNTRL_PAWN)==0)
              score -= (b1&15)*2;     
         }     
                  
                          
 


          if(p==PAWN){
             int half_passed = OpenLine(c,sq);
             int isolPawn = 0;
             int badPawn = 0;
             int passed_pawn = 0;

             if( half_passed ) passed_pawn = PassedPawn(c,sq);


             //   
             if((b_atk0[sq+dP]&CNTRL_PAWN)==0) score -= 1;


             // DOUBLE
             if(pawn_column_cnt[x] > 1){
                 score -= PDOUBLE;
                 if( ++cntDoublePawn > 1 ) score -= 1;
             }

             //BACKWARD
             if(BackPawn( c, sq )){
               score -= PBACK;
               if(half_passed) score -= 1;
               badPawn = 1;
               //ISOLATE
               if( (x==0 || pawn_column_cnt[x-1]==0) &&
                 (x==7 || pawn_column_cnt[x+1]==0) ) {
                isolCnt++;
                isolPawn = 1;
               }
             }
             //BLOCKED
             if( g.color[sq+dP] != NEUTRAL ){
               score -= PBLOCK;
               if(c==WHITE){
                 if(sq==E2 || sq==D2) score -= 2;
               }else{
                 if(sq==E7 || sq==D7) score -= 2;
               }
               if( g.color[sq+dP]==c1 ){
                  score -= PBLOCK_OP[x];
                  if( passed_pawn ) score -= PBLOCK_PASSED;
               }
             }

             //ATTACK BAD PAWN
//             if( badPawn ){
 //               score -=  (b1 & 15) * 3;    // counter attack
 //            }

             //PASSED
             if( passed_pawn )
             {
               int u = sq+dP;
               int op_blocked_atk = 0;
               while( (unsigned)u < 64 ){
                 if(b_atk1[u] & 15)  op_blocked_atk |= 1; //count. attacks
                 if( g.color[u]==c1 )  op_blocked_atk |= 2;
                 u += dP;
               }

               score += 22   // 
                      +  (0.8 + __stage*0.5) * (PASSED_PAWN_ROW[c][y]) // 
                      -  op_blocked_atk; // max 3
               if(++cntPassed == 1)  score += 4;   // 


               if(RookAccompaniesPassedPawn(c,sq))
                score += RAPP;


               // ****   ******
               passedPawnColMask |= (1 << x);
               if( (((1<<x) << 1) & passedPawnColMask)  |
                   (((1<<x) >> 1) & passedPawnColMask) )
               {
                  score += 2;                          //  

               }
               if( (b_atk0[sq] & CNTRL_PAWN)  ||       //  
                   (b_atk0[sq+dP] & CNTRL_PAWN) )
                   if( opponent_board[ y ] == c ) //  
                     score += 8;

              ///  **** end game ****
              if(stage==0  &&  g.cnt[c1][KNIGHT]+g.cnt[c1][BISHOP]+g.cnt[c1][ROOK]==0)
              {

                 int destSq = c==WHITE? x :  8*7 + x;
                 if( Distance(sq,destSq) < Distance(destSq, opKingSq) ){ //  - 100% !!!
                    score += PASSED_PAWN_ROW[c][y] << 3; //
                 }else{
                   int rookPawn = (x==0) | (x==7);//
                   //     
                   if( b_atk0[sq+dP] & CNTRL_KING ){
                       score += 2;
                       //     -  (  )
                       if(!rookPawn && opponent_board[y]==c ) //  
                         score += 8;
                   }
                   //      
                   //
                   if( rookPawn ){
                     //score -= 7-Distance(destSq, opKingSq);
                     if(  opKingSq == destSq  )  score -= 4;
                   }
                 }
              }// one king

              }//passed

             // - ->   --     -  
             if(half_passed  & !isolPawn){
                    score += 1;

                    if(stage < 0.5)
                        score += (0.5-stage) * HALF_PASSED_PAWN_ROW[c][y] ;
             }


           }else if(p==KING){
           
//////////////
    const int king_file[] = {0,0,3,4,4,3,0,0};
    double st = stage;
    int tmp = score_table[c][KING][sq];
    score -= tmp; // remove exist score
    if(g.cnt[c][PAWN]==0) st = 0; //    
    if(st > 0.5) tmp += king_file[COLUMN(sq)];
    score +=  st*(-tmp)  +  (1.0 - st)*tmp;
/////////////

              score += KingSFTY(c);

              if( g.cnt[c^1][QUEEN] )
               if( KingAroundBadMainDiag(c) )
               {
                 score -= 4;
               }

           }else{// other pieces

               if(p==ROOK){
                 score += Mobile_Pin_XRay(sq,p,c,b_atk0,b_atk1) +
                          main_line[x];;
              //   score += (14 - Taxi(sq, opKingSq))*__stageNB/2;
                 if(OpenLine(c,sq)){
                     score +=  RHOF;
                     if(OperatingLine(sq))
                         score += ROF;
                 }
                 if(b0 & CNTRL_ROOK) score += 2;

                 switch( rank[c][y] ){
                     case 0: score += 2; break; //    
                     case 6: if( HasPawn7Rank(c1) || rank[c1][ROW(opKingSq)] == 0)
                       {
                         rook7RankCnt++;
                         score += R7R + R7R_ZOOM * __stageNB;
                         if( rook7RankCnt > 1 ) score += 4;
                       }
                       break;
                     case 7:
                        //okThreat++;
                        score += 2;
                     break;
                   }

                 score += (8-g.cnt[c1][PAWN]) / 2; //  ,   

               }else if(p==BISHOP){
                 int t =  main_diag[sq];
                 score += t;
                 score += Mobile_Pin_XRay(sq,p,c,b_atk0,b_atk1);

                 if(g.cnt[c][BISHOP]>1) score += 8;
                 //
                 if(b0 & CNTRL_PAWN)   // 
                  if(pawn_column_cnt[x]==0) //  
                   if(g.cnt[c][ROOK])  //   ??
                     score += 2;  // - ,      


               }else if(p==QUEEN){
                 int threat;
                 score += Mobile_Pin_XRay(sq,p,c,b_atk0,b_atk1);
                 if(b0 & CNTRL_ROOK){
                    score += 1;
                    if( rank[c][y] == 6 && rook7RankCnt )
                      score += 2;
                 }

                 if(b_atk0[sq] & CNTRL_BISHOP) score += 1;
                 if(Distance(sq, g.kingSq[c]) > 2)  score -= 1;
               
               // 1 variant
                 threat = 14-Taxi(sq,opKingSq);
                 score += (double)threat*0.8 + (double)threat*0.4*__stageNB;
               // 2 variant  
                // if(Distance(sq, opKingSq) <= 3)
                  //  score += 5;                                    
                
               }else if(p==KNIGHT){
                  if(b_atk0[sq] & CNTRL_KNIGHT) score += 1;
                  if( Distance(sq, opKingSq) <= 3) {
                     score +=  NNK;
                  }
                  //   -     ()
                  score -= (16 - g.cnt[WHITE][PAWN] +  g.cnt[BLACK][PAWN]) >> 1;
                 //
                 if(b0 & CNTRL_PAWN)   // 
                  if(pawn_column_cnt[x]==0){ //  
                    score += 1;
                    if(opponent_board[y]==c)
                    {
                      score += knight_forpost_column[x];
                    }
                  }
               }
           }
       }//end if sq != -1


       score -= isolCnt*isolCnt/2;
       score -= hungCnt*hungCnt/2;
       //  
       if(g.cnt[c][PAWN]==0) score -= 4;
       s[c] += score;
       
       
       
       
     }//for c


  }while(0);



  save_alpha = alpha; save_beta = beta; save_side = g.side; save_key = g.key;
  save_score = s[g.side] - s[g.xside];


  return save_score;

}


void  GenScoreTable(int p, int tbl[64], int c, int max, int op_king_sq){


   if(p != PAWN){
      int tmp[64],j,cnt,d,u,n1,max_val,sq;

      for(j = 0; j < 64; j++){
        int row = c==BLACK? ROW(j) : (7 - ROW(j));
        tbl[j] = row;
        tmp[j] = 0;
      }
      max_val = 0;
      for(cnt = 0; cnt < 3; cnt++){
        for(sq = 0; sq < 64; sq++)
          for(d = d_start[p]; d <= d_stop[p]; d++)
            for(n1 = MAP(sq)+ dir[d]; (n1&0x88)==0; n1 += dir[d]){
              u = UNMAP(n1);
              tmp[sq] += tbl[u];
              if(tmp[sq] > max_val) max_val = tmp[sq];
              if(!SWEEP(p)) break;
            }
        //copy
        for(j = 0; j < 64; j++)
          tbl[j] = tmp[j];
      }
      //zoom result
      for(j = 0; j < 64; j++)
        tbl[j] = (unsigned)tbl[j] * max / max_val;

   }else{ //PAWN
    int CntAtkPromoteSq(int sq, int c);
    int PromotePathLen(int sq, int c);
    double temp[64], max_val = 0;
    int j;

    for(j = 0; j < 64; j++)
      if(ROW(j) != 0  &&  ROW(j) != 7){
        temp[j] = (double)CntAtkPromoteSq(j,c) /
                  PromotePathLen(j,c);
        if(temp[j] > max_val) max_val = temp[j];
      }else temp[j] = 0;
    for(j = 0; j < 64; j++)
      tbl[j] = temp[j] * max / max_val;
    //  
    if(c == WHITE){
       tbl[D4] += 2;
       tbl[E4] += 2;
       tbl[D3] += 1;
       tbl[E3] += 1;
       tbl[C3] += 1;
       tbl[C4] += 1;
       tbl[F4] += 1;

    }else{
       tbl[D5] += 2;
       tbl[E5] += 2;
       tbl[D6] += 1;
       tbl[E6] += 1;
       tbl[C6] += 1;
       tbl[C5] += 1;
       tbl[F5] += 1;

    }

   }

}


int CntAtkPromoteSq(int sq, int c){
 int x1,y1,x2,y2,dY;
  //left
  x1 = COLUMN(sq); y1 = ROW(sq);
  dY = c==WHITE? -1:1;
  do{
    x1 -= 1; y1 += dY;
  }while(x1 >= 0 && x1 <= 7 && y1 >= 0 && y1 <= 7);

  //right
  x2 = COLUMN(sq); y2 = ROW(sq);
  dY = c==WHITE? -1:1;
  do{
    x2 += 1; y2 += dY;
  }while(x2 >= 0 && x2 <= 7 && y2 >= 0 && y2 <= 7);

  return x2 - x1 + 1;

}

int PromotePathLen(int sq, int c){
  if(c==WHITE) return ROW(sq);
  else return 7 - ROW(sq);
}

void PrintArray(FILE*f,int tbl[64], int p, int c){
  char *chp[] = {"nopiece","pawn","knight","bishop",
                 "rook","queen","king"};
  char *chc[] = {"white","black"};
  int x,y;

  fprintf(f, "%s %s\n", chc[c], chp[p]);

  for(y = 0; y < 8; y++){
    for(x = 0; x < 8; x++){
      int j = y*8+x;

      fprintf(f,"%10d,",tbl[j]);
    }
    fprintf(f,"\n");
  }


}


void GenerateAttackes(void){
  int c0,i,sq,p,cntrl,j,*bit_atk,dP,n0,n1,u,d,p1;
    memset(p_attack,0,sizeof(struct Atk_Type)*2);
    for(c0 = WHITE; c0 <= BLACK; c0++){
      //c1 = c0 ^ 1;
      dP = c0==WHITE? -8:8;
      bit_atk = p_attack[c0].bit_atk;
      for(i = g.start[c0]; i <= g.stop[c0]; i++)
        if(  (sq = g.list[i]) != -1 ){
          p = g.pos[sq];
          cntrl = piece_cntrl[p];
          n0 = MAP(sq);
          if(p==PAWN){
             n1 = n0 + dP + dP - 1;
             if( (n1&0x88)==0 ){
               u = UNMAP(n1);
               bit_atk[u] = ++bit_atk[u] | cntrl;
             }
             n1 += 2;
             if( (n1&0x88)==0 ){
               u = UNMAP(n1);
               bit_atk[u] = ++bit_atk[u] | cntrl;
             }
          }else if(SWEEP(p)){//other pieces
             extern int atk[16*16],dir_atk[16*16];
             extern const int center;

             for(j = d_start[p]; j <= d_stop[p]; j++){
                d = dir[j];
                for(n1 = n0+d; (n1&0x88)==0; n1 += d){
                  u = UNMAP(n1);
                  bit_atk[u] = ++bit_atk[u] | cntrl;
                  p1 = g.pos[u];
                  if(p1==0) continue;
                  if( SWEEP(p1) && g.color[u]==c0 && (atk[n1-n0+center] & (1<<p1))){
                      for(n1 = n1+d; (n1&0x88)==0; n1 += d){
                         u = UNMAP(n1);
                         ++bit_atk[u];
                         if(g.pos[u]) break;
                      }
                  } //enif
                  break; //out for n1
                }//for n1
             }//for j
          }else{//KNIGHT, KING

             for(j = d_start[p]; j <= d_stop[p]; j++)
               if( ((n1 = n0+dir[j]) & 0x88) == 0 ){
                  u = UNMAP(n1);
                  bit_atk[u] = ++bit_atk[u] | cntrl;
               }
          }

        }
    }//for c
}



/*
   
*/

int KingSFTY(int c){
   int sq = g.kingSq[c], c1 = c^1, n0 = MAP(sq),
       *a1 = p_attack[c1].bit_atk,
       n1,d,u,j,cntMoveKing=0,
       cntNextCheck=0, attackKingSq =0,
       score = 0,
       *op_piece_cnt = g.cnt[c1];
       int one_king = g.mtl[c] == VALUE_K;
       int w;



    // COUNTER OF CHECK
 if(!one_king)
 {
  //// 0001    
    
   if(op_piece_cnt[BISHOP] | op_piece_cnt[QUEEN])
    for(j = d_start[BISHOP]; j <= d_stop[BISHOP]; j++){
     d = dir[j];
     for(n1 = n0+d; (n1&0x88)==0; n1+=d){
       u = UNMAP(n1);
       if(g.color[u] != c1){
         if(w=(a1[u] & (CNTRL_BISHOP|CNTRL_QUEEN)))
         {  
           cntNextCheck++;
            if( w &= w-1 ) cntNextCheck++;
         }
       }
       if(g.color[u] != NEUTRAL) break;
     }
    }

   if(op_piece_cnt[ROOK] | op_piece_cnt[QUEEN])
    for(j = d_start[ROOK]; j <= d_stop[ROOK]; j++){
     d = dir[j];
     for(n1 = n0+d; (n1&0x88)==0; n1+=d){
       u = UNMAP(n1);
       if(g.color[u] != c1){
         if(w=(a1[u] & (CNTRL_ROOK|CNTRL_QUEEN)))
         {  
           cntNextCheck++;
            if( w &= w-1 ) cntNextCheck++;
         }
       }
       if(g.color[u] != NEUTRAL) break;
     }
    }

  if(op_piece_cnt[KNIGHT])
    for(j = d_start[KNIGHT]; j <= d_stop[KNIGHT]; j++){
     if( ((n1 = n0+dir[j])&0x88)==0 ){
       u = UNMAP(n1);
       if(g.color[u] != c1)
         if(w=(a1[u] & (CNTRL_KNIGHT)))
         {  
           cntNextCheck++;
            if( w &= w-1 ) cntNextCheck++;
         }
     }
    }
//// 0001
 }//endif

   // counter of legal moves & atk. kins square
   
   for(j = d_start[KING]; j <= d_stop[KING]; j++)
    if( ((n1 = n0+dir[j])&0x88)==0 ){
      u = UNMAP(n1);
      if( a1[u] )
      {
        attackKingSq += a1[u] & 15;
      }else
          if(g.pos[u]==0) cntMoveKing++;
    }


   score_tree[ply].king_can_check[c] = cntNextCheck;

   switch(cntMoveKing){
     case 0: score -= 4; if( cntNextCheck ) score -= 4; break;
     case 1: score -= 2; break;
     case 2: score -= 1; break;
   }
   
   if(one_king) // end game ---> can mated?
   {
     int can_mated = op_piece_cnt[KNIGHT]*2 + 
                     op_piece_cnt[BISHOP]*3 +
                     op_piece_cnt[ROOK]*6 + 
                     op_piece_cnt[QUEEN]*10;
      if(can_mated >= 6 ){
        score -= 20000;
      }else if(can_mated == 5) // bishop + knight mated
      {
        score -= 200;
        if(cntMoveKing < 3) score -= 12;
        if(HasWhiteBishop(c^1))
        {
          score += 2*min( Distance(A8,sq), Distance(H1,sq));
        }else{
          score += 2*min( Distance(A1,sq), Distance(H8,sq));        
        }
        if((ROW(sq)==0) | 
           (ROW(sq)==7) | 
           (COLUMN(sq)==0) | 
           (COLUMN(sq)==7))
           score -= 10;
      }else if(can_mated==4){ //knight + knight
      
          score += 400; // not mated
          if(cntMoveKing < 3) score -= 12; // try??
      }else
          score += 20000; // not mated exactly
      
      return score;
   }

 
 

// if(cntNextCheck)
 //  score -= 4;

 if(op_piece_cnt[QUEEN]==0)
 {
   if(cntNextCheck)
   {
      score -= 4;
      if(cntNextCheck > 1) score -= 1;
   }
 }else{ // opponent hash queen
 
  if( PawnFriendKing(c,sq) == 0 ){
       score -= 8;
  }
  if(cntNextCheck)
  {
    
    // score -= 6;
    //if(cntNextCheck > 1) score -= 2;
     score -= 5  + cntNextCheck*cntNextCheck;
  }  
  if((op_piece_cnt[ROOK] | op_piece_cnt[BISHOP] | op_piece_cnt[KNIGHT]) )
  {
   /// 001
    ///f(cntNextCheck  && cntMoveKing <= 1)
     //    score -= 4;
    // score -= attackKingSq;


 ////////////////
 // king screen pawn struct
 do{
   int u;
   int notIsolCnt = 0;
   int *p_file_cnt = g.pawn_column_cnt[c];
   
   for( u = COLUMN(sq)-1; u <= COLUMN(sq)+1; u += 1)
    if(u >= 0 && u < 8)
    {
      if( p_file_cnt[u]==0 ){ // open line
        if( u == COLUMN(sq) ) score -= 8;
        else score -= 4;
      }else{
          if( (u==0 || p_file_cnt[u-1]==0) &&
              (u==7 || p_file_cnt[u+1]==0) )  // isol pawn
          {
          }else notIsolCnt++;
      }
    }

  if( notIsolCnt == 0 ) score -= 16;
  else if( notIsolCnt == 1 ) score -= 4;

 }while(0);





   //     
   if(!g.casling[c]){
     int LeftCastlEnable( int c );
     int RightCastlEnable( int c );
     int s = 0;
     int l = LeftCastlEnable(c);
     int r = RightCastlEnable(c);
     if(l==0) s -= 2;
     if(r==0) s -= 4;
     if(l + r == 0) s -= 14;
     score += s;
   }else{
      if( COLUMN( g.kingSq[c] ) >  4 ) //o-o
      {
         score += 10;

         //  
         /*
            . k . . . . k .
            p p . . . . p p  <---main pawns for castling
            . . . . . . . .
            . . . . . . . .
            . . . . . . . .
            . . . . . . . . 
            p p . . . . p p  <---main pawns for castling
            . k . . . . k .                                                                                    
         */
         if(c==WHITE){
             if( (g.pos[G2]==PAWN) & (g.color[G2]==WHITE))
             {
                 score += 2;
               if((g.pos[H2]==PAWN) & (g.color[H2]==WHITE))
                 score += 2;
               else if((g.pos[H3]==PAWN) & (g.color[H3]==WHITE))
                 score += 1;
             }
         }else{
             if( (g.pos[G7]==PAWN) & (g.color[G7]==BLACK))
             {
                 score += 2;
               if((g.pos[H7]==PAWN) & (g.color[H7]==BLACK))
                 score += 2;
               else if((g.pos[H6]==PAWN) & (g.color[H6]==BLACK))
                 score += 1;
             }
         }

      }else{
         score += 2;// o-o-o
         if(c==WHITE){
             if( (g.pos[B2]==PAWN) & (g.color[B2]==WHITE))
             {
                 score += 2;
               if((g.pos[A2]==PAWN) & (g.color[A2]==WHITE))
                 score += 2;
               else if((g.pos[A3]==PAWN) & (g.color[A3]==WHITE))
                 score += 1;
             }
         }else{
             if( (g.pos[B7]==PAWN) & (g.color[B7]==BLACK))
             {
                 score += 2;
               if((g.pos[A7]==PAWN) & (g.color[A7]==BLACK))
                 score += 2;
               else if((g.pos[A6]==PAWN) & (g.color[A6]==BLACK))
                 score += 1;

             }
         }
      }
   }
///////
  /// 001
 }
} // end check safety  HAS QUEEN

 
   return score;
}


int Trapped(int sq, int p, int c, int* a0, int* a1){
   int  n0 = MAP(sq),n1,u,j,d;
   int mask = less_piece_cntrl[p];
   int not_sweep = !SWEEP(p);

   for(j = d_start[p]; j <= d_stop[p]; j++){
     d = dir[j];
     for(n1 = n0+d; (n1&0x88)==0; n1+=d){
        int t;
        u = UNMAP(n1);
        t = g.pos[u];
        if(t==0){
          if( a1[u] && (a0[u]==0 || (mask & a1[u])) )
          {
             //HUNG
          }else return 0;
        }
        if(t | not_sweep) break;
     }
   }
   return 1;
}


/*
  ,      
*/

int  Mobile_Pin_XRay(int sq, int p, int c, int* a0, int* a1){
   int score = 0, mobile = 0, n0 = MAP(sq),n1,u,sq1,j,d;

   for(j = d_start[p]; j <= d_stop[p]; j++){
     d = dir[j];
     for(n1 = n0+d; (n1&0x88)==0; n1+=d){
        u = UNMAP(n1);
        if(g.pos[u]==0){
            mobile++;
            continue;
        }else{
           if(g.pos[u]==PAWN || g.pos[u]==KING) break;
           sq1 = u;
           for(n1 = n1+d; (n1&0x88)==0; n1+=d){
              u = UNMAP(n1);
              if(g.pos[u]==0) continue;
              if(g.color[u]== (c^1)){
                  if(comp_value[g.pos[u]] > comp_value[p]  ||
                     a1[u]==0){

                     if(g.color[sq1]==c){
                        score += XRAY;
                        if(g.pos[u]==KING) score += 4;// like next check
                        score_tree[ply].king_can_check[c^1]++;
                     }else{
                        score += PIN;
                        if(g.pos[u]==KING){
                            score += 2; // 
                            //if(a0[sq1] & CNTRL_PAWN) score += 4;//  
                       }
                     }
                  }
              }
              break;
           }//for n1
           break;
        } //if not empty square
      }//for n1
   }//for j
   
 

   if(mobile==0) score -= 3;
   else if(mobile < 3) score += 1;
   else {
      score += 3;
      if( p != QUEEN )
        if( p >= 6 ) score += 2;
   }
   //if(p==QUEEN) return score + mobile/4;
   //return score + mobile/3;
   
   return score;

}

