/* 
   @author Korniloff E.
     2007
*/


#ifndef _LINUX
#include <windows.h>
#else
#include <unistd.h>
#include <sys/mman.h>
#include <errno.h>
typedef unsigned short int HWND;
#define PAGE_READWRITE PROT_READ
#define ERROR_ALREADY_EXISTS (183)
#define CreateFileMapping mmap
#endif

#include <math.h>
#include "chess.h"
#include "learn.h"


extern int Search_Only_Mtl;
//#define WHITE_SQ(sq)   (((63-sq)&~7) +  (7 - ((63-sq)&7)))
int ActualSq(int sq, int c){
   if(c == BLACK) return sq;
   sq = 63 - sq;

   return   (sq&~7) +  ( 7 - (sq&7));

}
int white_sq[64];
const char *chb[] = {
   "a8","b8","c8","d8","e8","f8","g8","h8",
   "a7","b7","c7","d7","e7","f7","g7","h7",
   "a6","b6","c6","d6","e6","f6","g6","h6",
   "a5","b5","c5","d5","e5","f5","g5","h5",
   "a4","b4","c4","d4","e4","f4","g4","h4",
   "a3","b3","c3","d3","e3","f3","g3","h3",
   "a2","b2","c2","d2","e2","f2","g2","h2",
   "a1","b1","c1","d1","e1","f1","g1","h1"
};


int StrToSq(char *s){
  int j;

  for(j = 0; j < 64; j++)
   if(chb[j][0]==s[0] && chb[j][1]==s[1]) return j;

  return -1;
}

void InitNewGame(void){
   int InitGame(char *);
   InitGame(START_POS_FEN);

}


int InitGame(char *epd){
  const int hlat[4][21] =
  {
    {'p','n','b','r','q','k',  'P','N','B','R','Q','K',  '1','2','3','4','5','6','7','8',  '/'},
    { BLACK,BLACK,BLACK,BLACK,BLACK,BLACK, WHITE,WHITE,WHITE,WHITE,WHITE,WHITE,  0,  0,  0,  0,  0,  0,  0,  0,   0},
    {PAWN,KNIGHT,BISHOP,ROOK,QUEEN,KING, PAWN,KNIGHT,BISHOP,ROOK,QUEEN,KING, 0,0,0,0,0,0,0,0, 0 },
    { 1,  1,  1,  1,  1,  1,    1,  1,  1,  1,  1,  1,    1,  2,  3,  4,  5,  6,  7,  8,    0}
  };
  char buf[256];
  char *lex;
  //int sq,p,c,k,j;
  int j,sq;
  void InsertPiece(int p, int c, int sq, int i);
  struct Game_Type save;
  int pos[2][64];

   if(epd==NULL) return 0;

   save = g;
   //CLEAR GAME
   memset(&g,0,sizeof(struct Game_Type));
   for(j = 0; j < 32; j++) g.list[j] = -1;
   for(j = 0; j < 64; j++){
      g.index[j] = -1;
      g.color[j] = NEUTRAL;
   }
   g.side = WHITE; g.xside = BLACK;

   memset(pos, 0, sizeof(pos));

   //READ POSITION
   strcpy(buf,epd);
   lex = strtok(buf," \n");
   if(!lex) goto error;

/////////????
   g.start[BLACK] = 16; g.stop[BLACK] = 15;
   g.start[WHITE] = 0; g.stop[WHITE] = -1;
   sq = 0;
   while(*lex){
     int ch = *lex++;
     if(ch=='/') ;
     else if(isdigit(ch)) sq += ch - '0';
     else{
       int c = islower(ch)? BLACK : WHITE;
       int p = NOPIECE;
       ch = tolower(ch);
       if(ch=='p') p = PAWN;
       else if(ch=='n') p = KNIGHT;
       else if(ch=='b') p = BISHOP;
       else if(ch=='r') p = ROOK;
       else if(ch=='q') p = QUEEN;
       else if(ch=='k') p = KING;
       else goto error;
       do{
           int i = ++g.stop[c];
           if(i >= (16<<c)) goto error;
           InsertPiece(p, c, sq, i);
          // g.rep_piece_table[i][sq] = 1;
       }while(0);
       sq++;
     }
   }
   if(g.cnt[WHITE][KING] != 1  ||  g.cnt[BLACK][KING] != 1) goto error;


///////????

   //READ SIDE
   lex = strtok(NULL," \n");
   if(!lex) goto success;

   if(lex[1] != '\0') goto error;
   if(*lex=='-');
   else if(*lex=='w'){
      g.side = WHITE; g.xside = BLACK;
   }else if(*lex=='b'){
      g.side = BLACK; g.xside = WHITE;
   }else goto error;



   //READ CASTLING ENABLE
   lex = strtok(NULL," \n");
   if(!lex) goto success;

   if(strcmp(lex,"-")==0);
   else
    while(*lex){
     switch(*lex){
       case 'Q':  g.castl_enable |= WHITE_LEFT_CASTL_ENABLE; break;
       case 'K':  g.castl_enable |= WHITE_RIGHT_CASTL_ENABLE; break;
       case 'q':  g.castl_enable |= BLACK_LEFT_CASTL_ENABLE; break;
       case 'k':  g.castl_enable |= BLACK_RIGHT_CASTL_ENABLE; break;
       default: goto error;
     }
     lex++;
    }

   //READ LAST MOVE SQ (for en passant only)
   lex = strtok(NULL," \n");
   if(!lex) goto success;

   if(strcmp(lex,"-")==0) goto success;
   if((sq=StrToSq(lex)) == -1) goto error;
   if(g.side==WHITE &&  ROW(sq) != 2) goto error;
   if(g.side==BLACK && ROW(sq) != 5) goto error;
   if(g.pos[sq - dP[g.side]] != PAWN  ||  g.color[sq - dP[g.side]] != g.xside ||
      g.pos[sq] != NOPIECE || g.pos[sq+dP[g.side]] != NOPIECE)
     goto error;
   g.__en_pass_sq = sq;


success:
   do{
     extern time_t limit_time,ext_time;

     limit_time = 7.5*100;
	 ext_time = 3*limit_time;
   }while(0);
  /*
   do{
     void PrintPList(char *fname);
     PrintPList("plist.txt");
   }while(0);
   */
   strcpy(g.fen, epd);

   return 1;

error:
  g = save;
  return 0;
}



void PrintPList(char *fname){
 int j;
 FILE *f = fopen(fname,"w");
 if(f)
 {
  for(j = 0; j < 32; j++){
   int sq = g.list[j];
   if(sq != -1){
     fprintf(f," piece %3d, color %3d, sq %3d, index %3d\n",
               g.pos[sq], g.color[sq], sq, j);

   }
  }
  fclose(f);
 }

}


int StrToMove(char *s, Move *ret_mv){
   int len = strlen(s), from, to, find = 0;

   if(len >= 4 && (from = StrToSq(s)) != -1  &&
      (to = StrToSq(s+2)) != -1)   find = 1;
   else if(strcmp(s,"o-o")==0){
       if(g.side==WHITE){
          from = E1; to = G1;
       }else{
          from = E8; to = G8;
       }
       find = 1;
   }else if(strcmp(s,"o-o-o")==0){
       if(g.side==WHITE){
          from = E1; to = C1;
       }else{
          from = E8; to = C8;
       }
       find = 1;
   }


   if(find){
      int LegalMoveList(void);
      int j;
      Move mv;

      LegalMoveList();
      for(j = 0; j < treeCnt[1]; j++){
        Move mv = tree[j];
        if(FROM(mv)==from && TO(mv)==to){
           if(mv & PROMOTE){
              int new_p = QUEEN;
              if(len > 4)
                switch(tolower(s[4])){
                  case 'q':   new_p = QUEEN; break;
                  case 'n':   new_p = KNIGHT; break;
                  case 'b':   new_p = BISHOP; break;
                  case 'r':   new_p = ROOK; break;
                }
                mv = (mv & ~(7<<12)) | (new_p << 12);
            }
           *ret_mv = mv;
           return 1;
        }
      }//
   }
   return 0;
}

int InsertMoveInGame(Move mv){
   unsigned ControlSumm(void *data, unsigned sz);
   if(g.game_cnt < MAX_GAME){
        
       Node *learn_node = SearchCurrPos();
     /*  
       unsigned temp = ControlSumm(&g,sizeof(struct Game_Type));
       MakeMove(mv);
       UnMakeMove(mv);
       assert(temp==ControlSumm(&g,sizeof(struct Game_Type)));
     */
       ply = 0;  
       MakeMove(mv);
       ply = 0;
       g.key_list[g.game_cnt] = g.key; 
       g.white_key_list[g.game_cnt] = g.side_key[WHITE];      
       g.black_key_list[g.game_cnt] = g.side_key[BLACK];             
       g.game_list[g.game_cnt] = mv;
       g.learn_node_list[g.game_cnt] = learn_node; 
       g.game_max = g.game_cnt = g.game_cnt + 1;
       
       return 1;
   }
   return 0;

}

char *MoveToStr(Move mv, char* str){
   sprintf(str,"%s%s", chb[FROM(mv)], chb[TO(mv)]);
   if(mv & PROMOTE){
     char *chp[] = {"","p","n","b","r","q","k"};
     strcat(str, chp[ PIECE(mv) ]);
   }
   return str;
}





void InitAllVar(void){
 //  void LearnClear(void);
   void HashInit(void);
   void InitEvaluate(void);
   void InitBitMask(void);
 //  LearnClear();
   InitBitMask();
   InitAtk();
   HashInit();
   InitEvaluate();

   do{
     int j;
     for(j = 0; j < 64; j++)
      white_sq[j] = ActualSq( j, WHITE );

   }while(0);
}


unsigned ControlSumm(void *data, unsigned sz){
   unsigned char *s = data;
   unsigned summ = 0, n = 0;

   while(n < sz){
     summ += *s;
     s++;
     n++;
   }
   return summ;
}

char* U64ToStr(char *str, U64 v){
  char *p,*s;
  p = s = str;
  do{
     *p++ = (char)('0' + v%10);

  }while( (v /= 10) > 0);
  *p-- = '\0';
  while(p > s){
    char temp = *p;
    *p = *s;
    *s = temp;
    s++; p--;
  }
  return str;
}


double  Round( double s ){
   
   double  integer;
   
   if( modf( s, &integer ) > 0.5 ) return ceil( s );
   else return floor( s );

}


double Int( double s ) {  return Round(s); }
double Fract( double s )  {
   return s - Round( s );
}


int MtlScore(int score){
 if( score < -INF + 100 )
   return -INF;
 else if (  score > INF - 100)
   return INF;
 else {
  int st  = score % 1000;
  int mtl = score / 1000 * 1000;
  if( score > 0 ){
     if( st > 500 ){ // -
        mtl += 1000;     //998
        st = st - 1000;   //-2
     }
  }else{
     if( st < -500 ){  // -998
        mtl -= 1000;
        st = 1000 + st;   //+2
     }
  }
//  st  = score - mtl;
  return mtl;
 }
}

int StScore(int score){
 if( score < -INF + 100 )
   return -INF;
 else if (  score > INF - 100)
   return INF;
 else {
  int st  = score % 1000;
  int mtl = score / 1000 * 1000;
  if( score > 0 ){
     if( st > 500 ){ // -
        mtl += 1000;     //998
        st = st - 1000;   //-2
     }
  }else{
     if( st < -500 ){  // -998       1   ,  > 0
        mtl -= 1000;
        st = 1000 + st;   //+2
     }
  }
//  st  = score - mtl;
  return st;
 }

}




 void PrintSearchStatus(int ply, int score, int time, U64 nodes, Move *pv){
  char str_nodes[128], str_move[64], output[1024];
  int    j;
  unsigned move_sec = 0;

  if(time/100 > 0)
	  move_sec = nodes/ (time/100);

  score = (int64_t)score * 100 / VALUE_P;
  sprintf(output, "%2d %4d %4d %10s ",ply, score , time, U64ToStr(str_nodes,move_sec));
  for(j = 0; j < 14 && pv[j]; j++) {
    strcat(output, MoveToStr(pv[j],str_move));
    strcat(output," ");
  }

  printf("%s\n", output);
}

FullMove* FullMoveInit(FullMove *p, int depth, int score, int time, Move mv,
                       Move line[] )
{
  p->score = score;
  p->depth = depth;
  p->time  = time;
  p->mv    = mv;
  if(!line) p->line[0] = 0;
  else{
    int j;
    for( j = 0; j < FullMoveMaxPvLine-1 && line[j]; j++)
     p->line[j] = line[j];
    p->line[j]  = 0;
  }
  return p;
}


int file_exist(char fname[])
{
  FILE *f = fopen(fname,"r");
  if(f==0) return 0;
  fclose(f);
  return 1;

}

int prog_is_run(void){
    HWND MemHnd;
    int flag;
    // file in memory
    MemHnd = CreateFileMapping((HWND)0xFFFFFFFF,
       0, PAGE_READWRITE, 0, 1024, "dummy");
   #ifndef _LINUX 
   flag = GetLastError() == ERROR_ALREADY_EXISTS;
   #else
   flag = errno == ERROR_ALREADY_EXISTS;
   #endif
    
    
 
   // CloseHandle(MemHnd);
    return flag;
  }
  /*
#else
  int prog_is_run(void) { 
  
     return 0; 
     
  } // write to do
#endif
*/
